#include <hmg.ch>
#include <GraphPlus.ch>
#include "hbclass.ch"

Class GraphPlus

   Data aData          INIT {}    // Graph data
   Data aLegends       INIT {}
   Data aCategories    INIT {}  // Horizontal Axis
   Data nGraphType     INIT GT_DEFAULT  // COLUMN
   Data cTitle         INIT '' // Graph Title
   Data cXTitle        INIT '' // Horizontal Title
   Data cYTitle        INIT '' // Vertical Title
   Data lXRotate       INIT .f. // xValues Rotate
   Data nHeight        INIT 600 // Graph Height
   Data nWidth         INIT 600 // Graph Width
   Data lShowLegends   INIT .f. // Show legends
   Data lDataLabels    INIT .f. // show data labels
   Data lHGrid         INIT .t. // Horizontal Grid
   Data lVGrid         INIT .t. // Vertical Grid
   Data lShowValues    INIT .t. // Show Value Axis
   Data lShowCategories INIT .t. // show categories
   Data nLegendPos     INIT LEGEND_ON_BOTTOM   // 0 Right 1 Bottom
   Data lOutBorder     INIT .t. // draw outside border
   Data aColors        INIT {}  // aColors
   Data aTitleFont     INIT { 'Arial', 12, .t., .f., .f., .f. }
   Data aLegendFont    INIT { 'Arial', 8, .f., .f., .f., .f. }
   Data aBackColor     INIT CLR_WHITE
   Data aTitleColor    INIT { 0, 0, 255 }
   Data aTextColor      INIT { 0, 0, 0 }
   Data aGridLineColor INIT { 200, 200, 200 }
   Data nLegendBoxSize INIT 10
   Data nLegendWidth   INIT 0
   Data nColorTheme    INIT 1
   Data nUpper         INIT 0
   Data nLower         INIT 0
   Data lCategoryBelow INIT .t.
   Data nRange         INIT 0
   Data nPixelperPoint INIT 0
   Data nLineWidth     INIT 2
   Data nZeroPos       INIT 0
   Data nVDivision     INIT 5
   Data nHDivision     INIT 0
   Data nUpperDivision INIT 0
   Data nLowerDivision INIT 0
   Data cPicture       INIT ''
   Data nCategoryRow   INIT 0
   Data nScaleItemRow  INIT 0
   Data nPieGap        INIT 3
   Data nBarGapRatio   INIT 0.2 // in percentage 20%
   Data nGapWidthRatio INIT 1  // in percentage 200%
   Data nImageHeight   INIT 600
   Data nImageWidth    INIT 600
   Data nPointSize     INIT 10
   Data lGraphRotated  INIT .f.
   Data nMaxScaleWidth INIT 0
   
   Data hBitMap        INIT nil // Bitmap to store Graph
   Data hDC            INIT nil 
   Data BTStruct       INIT nil  
   Data nTop           INIT 10
   Data nLeft          INIT 10
   Data nRight         INIT 590 
   Data nBottom        INIT 590
   Data aScale         INIT {}
   DATA aColorThemes INIT { { CLR_HAVELOCK_BLUE, CLR_WESTSIDE, CLR_LIGHT_GREY, CLR_ORANGE_YELLOW, CLR_PICTON_BLUE, CLR_APPLE }, ;
      { CLR_HAVELOCK_BLUE, CLR_LIGHT_GREY, CLR_PICTON_BLUE, CLR_DARKER_BLUE, CLR_DARK_GREY, CLR_ENDEAVOUR }, ;
      { CLR_WESTSIDE, CLR_ORANGE_YELLOW, CLR_APPLE, CLR_SADDLE_BROWN, CLR_LIGHT_BROWN, CLR_DELL }, ;
      { CLR_APPLE, CLR_PICTON_BLUE, CLR_ORANGE_YELLOW, CLR_DELL, CLR_ENDEAVOUR, CLR_LIGHT_BROWN }, ;
      { CLR_HAVELOCK_BLUE }, ;
      { CLR_WESTSIDE }, ;
      }
   Data lLighter       INIT .f.
   
   Method New 
   Method Draw
   Method Bitmap
   Method DrawColumnGraph
   Method DrawBarGraph
   Method DrawLineGraph
   Method DrawPieGraph
   Method DrawPointsGraph
   Method DrawFunnelGraph
   Method DrawLegends
   Method InitColors
   Method DrawXYTitles
   Method DrawCategories
   Method DrawVerticalScale
   Method DrawBarScale
   Method DrawHorizontalGrid
   Method DrawVerticalGrid
   Method SwitchRowsToColumns
   Method Save
   Method GraphData    SETGET
   Method GridData     SETGET
   Method Legends      SETGET
   Method Picture      SETGET
   Method Categories   SETGET
   Method GraphColors  SETGET
   Method GraphType    SETGET
   Method Title        SETGET
   Method XTitle       SETGET
   Method YTitle       SETGET
   Method Height       SETGET
   Method Width        SETGET
   Method ShowLegends  SETGET
   Method ShowLabels   SETGET
   Method ShowHGrid    SETGET
   Method ShowVGrid    SETGET
   Method ShowValues   SETGET
   Method LegendPos    SETGET
   Method ColorTheme   SETGET
   Method BackColor    SETGET
   Method BarGapRatio  SETGET
   Method BarGapWidthRatio  SETGET
   Method TitleFont    SETGET
   Method LegendFont   SETGET

End class   
   
METHOD New CLASS GraphPlus
RETURN self

METHOD GraphData( aData ) CLASS GraphPlus
   local i := 0, nLen := 0
   if PCount() == 0
      return ::aData
   else
      if valtype( aData[ 1 ] ) == 'A' // 2D array
         ::aData := aData
      else
         ::aData := { aData } // 1D - make it 2D
      endif      
   endif
   return Self
   
Method GridData( cFormName, cControlName ) Class GraphPlus
   local i := 0
   local j := 0
   local nRowCount := getproperty( cFormName, cControlName, 'ITEMCOUNT' )
   local nColumnCount := 0
   local aLineData := {}
   local aLine := {}
   if PCount() == 0
      return ::aData
   else
      asize( ::aData, 0 )
      asize( ::aLegends, 0 )
      asize( ::aCategories, 0 )
      nColumnCount := getproperty( cFormName, cControlName, 'COLUMNCOUNT' )
      for i := 1 to nColumnCount
         if i == 1
//            ::cYTitle := getproperty( cFormName, cControlName, 'COLUMNHEADER', i ) not meaningful!
         else   
            aadd( ::aCategories, getproperty( cFormName, cControlName, 'COLUMNHEADER', i ) )
         endif   
      next i
      if nRowCount > 0
         for i := 1 to nRowCount
            aLineData := getproperty( cFormName, cControlName, 'ITEM', i )
            asize( aLine, 0 )
            for j := 1 to nColumnCount
               if j == 1
                  aadd( ::aLegends, aLineData[ j ] )
               else   
                  aadd( aLine, val( alltrim( aLineData[ j ] ) ) )
               endif
            next j
            aadd( ::aData, aclone( aLine ) )
         next i
      endif   
   endif
return self   

Method Bitmap() Class GraphPlus
   return ::hBitMap   

METHOD GraphColors( aColors ) CLASS GraphPlus
   if PCount() == 0
      return ::aColors
   else
      ::aColors := aColors
   endif
   return Self
   
METHOD Legends( aLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::aLegends
   else
      ::aLegends := aLegends
   endif
   return Self
   
METHOD Categories( aCategories ) CLASS GraphPlus
   if PCount() == 0
      return ::aCategories
   else
      ::aCategories := aCategories
   endif
   return Self

METHOD BackColor( aBackColor ) CLASS GraphPlus
   if PCount() == 0
      return ::aBackColor
   else
      ::aBackColor := aBackColor
   endif
   return Self
   
METHOD GraphType( nGraphType ) CLASS GraphPlus
   if PCount() == 0
      return ::nGraphType
   else
      ::nGraphType := nGraphType
      if ::nGraphType == 5 .or. ::nGraphType == 6 // bar or funnel
         ::lGraphRotated := .t.
      else
         ::lGraphRotated := .f.
      endif
   endif
   return Self

METHOD Title( cTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cTitle
   else
      ::cTitle := cTitle
   endif
   return Self
   
METHOD XTitle( cXTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cXTitle
   else
      ::cXTitle := cXTitle
   endif
   return Self
      
METHOD YTitle( cYTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cYTitle
   else
      ::cYTitle := cYTitle
   endif
   return Self
   
METHOD BarGapRatio( nBarGapRatio ) CLASS GraphPlus
   if PCount() == 0
      return ::nBarGapRatio
   else
      ::nBarGapRatio := nBarGapRatio
   endif
   return Self

METHOD BarGapWidthRatio( nBarGapWidthRatio ) CLASS GraphPlus
   if PCount() == 0
      return ::nGapWidthRatio
   else
      ::nGapWidthRatio := nBarGapWidthRatio
   endif
   return Self
   
METHOD TitleFont( aTitleFont ) CLASS GraphPlus
   if PCount() == 0
      return ::aTitleFont
   else
      ::aTitleFont := aTitleFont
   endif
   return Self

METHOD LegendFont( aLegendFont ) CLASS GraphPlus
   if PCount() == 0
      return ::aLegendFont
   else
      ::aLegendFont := aLegendFont
   endif
   return Self
   
METHOD Height( nHeight ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageHeight
   else
      ::nImageHeight := nHeight
   endif
   return Self

METHOD Width( nWidth ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageWidth
   else
      ::nImageWidth := nWidth
   endif
   return Self

METHOD ShowLegends( lShowLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowLegends
   else
      ::lShowLegends := lShowLegends
   endif
   return Self

METHOD ShowLabels( lDataLabels ) CLASS GraphPlus
   if PCount() == 0
      return ::lDataLabels
   else
      ::lDataLabels := lDataLabels
   endif
   return Self

METHOD ShowValues( lShowValues ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowValues
   else
      ::lShowValues := lShowValues
   endif
   return Self

METHOD ShowHGrid( lHGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lHGrid
   else
      ::lHGrid := lHGrid
   endif
   return Self

METHOD ShowVGrid( lVGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lVGrid
   else
      ::lVGrid := lVGrid
   endif
   return Self
   
METHOD Picture( cPicture ) CLASS GraphPlus
   if PCount() == 0
      return ::cPicture
   else
      ::cPicture := cPicture
   endif
   return Self
   

Method LegendPos( nLegendPos ) CLASS GraphPlus
   if PCount() == 0
      return ::nLegendPos
   else
      ::nLegendPos := nLegendPos
   endif
   return Self

Method ColorTheme( nColorTheme ) CLASS GraphPlus
   if PCount() == 0
      return ::nColorTheme
   else
      ::nColorTheme := nColorTheme
   endif
   return Self

   
Method Draw CLASS GraphPlus
   local aSize := {}, nLowerUpperRatio := 0
   local nDistance := 0
   local nRow := 0, nPixels := 0
   local nCount := 0
   local aCategories := {}
   local nZero := 0
   local nMaxWidth := 0, nMaxHeight := 0
   local i := 0
   LOCAL nMax, nMin, nI, nJ
   LOCAL nDivision, cDivision, nDecimalLength, nDecimal, cDecimal
   if hmg_len( ::aData ) == 0
      // nothing to draw!
      msgstop( 'No Graph data!' ) 
      return nil
   endif
   ::nTop := 10
   ::nLeft := 10
   ::nHeight := ::nImageHeight
   ::nWidth := ::nImageWidth
   ::nBottom := ::nHeight - 10
   ::nRight := ::nWidth - 10
   // init bitmap
   ::hBitmap := BT_BitmapCreateNew ( ::nWidth, ::nHeight, ::aBackColor )
   ::hDC := BT_CreateDC( ::hBitmap, BT_HDC_BITMAP, @::BTStruct )
   ::aTextColor := GetNegativeColor ( ::aBackColor )
   if hmg_len( ::aLegends ) == 0
      for i := 1 to hmg_len( ::aData )
         aadd( ::aLegends, "Legend " + alltrim( str( i ) ) )
      next i
   endif   
   if hmg_len( ::aCategories ) == 0
      // categories not defined
      for i := 1 to hmg_len( ::aData )
         aadd( ::aCategories, 'Category ' + alltrim( str( i ) ) )
      next i
   endif
   // draw outside border
   if ::lOutBorder
      BT_DrawRectangle (::hDC, 0, 0, ::nWidth - 1 , ::nHeight - 1, ::aTextColor, 1 )
   endif   
   ::InitColors()    // init colors
   if ::nGraphType == 6 // Funnel Chart
      ::DrawFunnelGraph()
      BT_DeleteDC( ::BTstruct )
      return nil
   endif
   // draw legends
   if ::lShowLegends .and. hmg_len( ::aLegends ) > 0
      ::DrawLegends()
   endif   
   // draw title
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText ( ::hDC, ::nTop / 2, ( ::nWidth - if( ::lShowLegends .AND. ::nLegendPos < 1, ::nLegendWidth, 0 ) ) / 2, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      ::nTop := ::nTop + aSize[ 2 ]
   endif
   if ::nGraphType <> 4 // not pie
      ::DrawXYTitles()    // draw x/y titles
      // draw graph border
      && BT_DrawRectangle (::hDC, ::nTop, ::nLeft, ::nRight - ::nLeft , ::nBottom - ::nTop, ::aTextColor, 1 )
      
      // find max and min value
      nMax := 0
      nMin := 0
      FOR nJ := 1 TO HMG_LEN( ::aData )
         FOR nI :=1 TO HMG_LEN( ::aData[ nJ ] )
            nMax := Max( ::aData[nJ][nI], nMax )
            nMin := Min( ::aData[nJ][nI], nMin )
         NEXT nI
      NEXT nJ
      if nMax > 0
         ::nUpper := nMax
      else 
         ::nUpper := 0
      endif   
      if nMin < 0
         ::nLower := nMin
      else
         ::nLower := 0   
      endif
      ::nRange := ::nUpper - ::nLower
      if ::nUpper <> 0 .and. ::nLower <> 0
         if abs( ::nLower ) < ::nUpper 
            nLowerUpperRatio := abs( ::nLower ) / ::nRange
            ::nLowerDivision := ::nVDivision * nLowerUpperRatio
            if ::nLowerDivision - int( ::nLowerDivision ) > 0.0
               ::nLowerDivision := int( ::nLowerDivision ) + 1
               ::nUpperDivision := ::nVDivision - ::nLowerDivision
            endif   
            nDivision := ::nUpper / ::nUpperDivision
         else
            nLowerUpperRatio := abs( ::nUpper ) / ::nRange
            ::nUpperDivision := ::nVDivision * nLowerUpperRatio
            if ::nUpperDivision - int( ::nUpperDivision ) > 0.0
               ::nUpperDivision := int( ::nUpperDivision ) + 1
               ::nLowerDivision := ::nVDivision - ::nUpperDivision
            endif   
            nDivision := abs( ::nLower ) / ::nLowerDivision
         endif
      else
         nDivision := ::nRange / ::nVDivision
         if ::nUpper > 0
            ::nUpperDivision := ::nVDivision
            ::nLowerDivision := 0
         else
            ::nLowerDivision := ::nVDivision
            ::nUpperDivision := 0
         endif         
      endif         
      if int( nDivision ) > 0
         cDivision := alltrim( str( int( nDivision ), 14, 0 ) )
         if hmg_len( cDivision ) > 1 // if more than 1 digits round to power of 10s
            nDivision := ( val( left( cDivision, 2 ) ) + 1 ) * 10^ ( hmg_len( cDivision ) - 2 )
         else
            if nDivision > int( nDivision ) 
               nDivision++
            endif
         endif
      else //
         cDivision := transform( nDivision, '9.9999999' )
         nDecimalLength := 8 - hmg_len( alltrim( str( val( right( cDivision, 7 ) ) ) ) )
         nDecimal := val( right( cDivision, 7 ) )
         cDecimal := alltrim( str( nDecimal ) )
         nDivision := (val( alltrim( left( cDecimal, 1 ) )) + 1 ) /(10 ^nDecimalLength)
      endif
      ::nRange := ::nVDivision * nDivision 
      if hmg_len( ::cPicture ) == 0
         ::cPicture := replicate( '9', max( hmg_len( alltrim( str( ::nUpper ) ) ), hmg_len( alltrim( str( ::nLower ) ) ) ) )
      endif   
      // find scales
      asize( ::aScale, 0 )
      nDistance := ::nRange / ::nVDivision
      for i := ::nUpperDivision to 1 step -1
         aadd( ::aScale, alltrim( transform( nDistance * i, ::cPicture ) ) )
      next i
      aadd( ::aScale, alltrim( transform( 0, ::cPicture ) ) )
      nZero := hmg_len( ::aScale )
      for i := 1 to ::nLowerDivision
         aadd( ::aScale, '-' + alltrim( transform( nDistance * i, ::cPicture ) ) )
      next i
      // find maxscaleitemwidth
      nMaxWidth := 0
      for i := 1 to hmg_len( ::aScale )
         aSize := BT_DrawTextSize (::hDC, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      next i
      ::nMaxScaleWidth := nMaxWidth
      // draw all items other than graph
      ::nTop := ::nTop + 20
      ::nBottom := ::nBottom - 20
      if ::lShowCategories 
         ::DrawCategories()
      endif
      // scale
      if ::lShowValues 
         if .not. ::lGraphRotated // not bar
            ::DrawVerticalScale()
         else
            ::DrawBarScale()
         endif   
      endif
      ::nRight := ::nRight - 5
      ::nLeft := ::nLeft + 5
      // find zero position
      do case
      case nZero == 1 // all are negative!
         if .not. ::lGraphRotated //bar
            ::nZeroPos := ::nTop
         else
            ::nZeroPos := ::nRight
         endif
      case nZero == hmg_len( ::aScale ) // all are positive!
         if .not. ::lGraphRotated //other than bar
            ::nZeroPos := ::nBottom
         else
            ::nZeroPos := ::nLeft
         endif
      otherwise  // zero position in the middles
         if .not. ::lGraphRotated // other than bar
            ::nZeroPos := ::nTop + ( ( ::nBottom - ::nTop ) / ::nVDivision * ( nZero - 1 ) )
            ::nPixelperPoint := ( ::nBottom - ::nTop ) / ::nRange
         else // bar
            ::nZeroPos := ::nRight - ( ( ::nRight - ::nLeft ) / ::nVDivision * ( nZero - 1 ) )
            ::nPixelperPoint := ( ::nRight - ::nLeft ) / ::nRange
         endif
      endcase
      if ::lHGrid 
         ::DrawHorizontalGrid()
      endif
      if ::lVGrid 
         ::DrawVerticalGrid()
      endif
   endif   
   do case
   case ::nGraphType == 1 // Columns
      ::DrawColumnGraph()
   case ::nGraphType == 2 // Line
      ::DrawLineGraph()
   case ::nGraphType == 3 // Points
      ::DrawPointsGraph()
   case ::nGraphType == 4 // Pie
      ::DrawPieGraph()
   case ::nGraphType == 5 // Bar
      ::DrawBarGraph()
   endcase   
   BT_DeleteDC( ::BTstruct )
   
return nil   

Method DrawColumnGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   LOCAL i, j, nBarHeight
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio 
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nGapWidthRatio
   nBarGap      := nBarWidth * ::nBarGapRatio
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow - nBarHeight, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3, nCol + (nBarWidth / 2 ), if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
         nCol := nCol + nBarWidth 
         if j < hmg_len( ::aData )
            nCol := nCol + nBarGap
         endif   
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawVerticalScale Class GraphPlus
   local nScaleItems := 0
   local nRow := 0, nCol := 0, nData := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nBarHeight := 0
   local nScaleItemHeight := 0
   local aSize := {}, i := 0
   local nMaxWidth := 0
   nScaleItems := hmg_len( ::aScale )
   if nScaleItems == 0
      return nil
   endif
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nScaleItems - 1 // equal space between scales
   nScaleItemHeight := nAvailableHeight / nTotalBarRatio 
   nRow := ::nTop
   for i := 1 to hmg_len( ::aScale )
      aSize := BT_DrawTextSize (::hDC, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
   next i  
   ::nLeft := ::nLeft + nMaxWidth
   for i := 1 to hmg_len( ::aScale )
      BT_DrawText ( ::hDC, nRow + 2, ::nLeft, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_BASELINE + BT_TEXT_RIGHT)
      nRow := nRow + nScaleItemHeight
   next i
return nil

Method DrawHorizontalGrid Class GraphPlus
   local i := 0
   local nHeight := ( ::nBottom - ::nTop ) / ::nVDivision
   local nRow := 0
   nRow := ::nTop
   if .not. ::lGraphRotated // not bar
      for i := 1 to hmg_len( ::aScale )
         BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
         nRow := nRow + nHeight
      next i
   else 
      nHeight := ( ::nBottom - ::nTop ) / hmg_len( ::aCategories )
      nRow := ::nTop
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
         nRow := nRow + nHeight
      next i
      BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
   endif   
return nil

Method DrawVerticalGrid Class GraphPlus
   local i := 0
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nWidth := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nBarWidth := 0
   local nCategoryWidth := 0
   if .not. ::lGraphRotated
      nCategories := hmg_len( ::aData[ 1 ] )
      if nCategories == 0
         return nil
      endif
      nAvailableWidth := ::nRight - ::nLeft
      nTotalBarRatio := nCategories // equal space between categories
      nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
      nCategoryWidth := nBarWidth 
      BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
      nCol := ::nLeft
      for i := 1 to nCategories
         nCol := nCol + nCategoryWidth
         BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
      next i
   else
      nWidth := ( ::nRight - ::nLeft ) / ( hmg_len( ::aScale ) - 1 )
      nCol := ::nLeft
      for i := 1 to hmg_len( ::aScale )
         BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
         nCol := nCol + nWidth
      next i
      BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
   endif   
return nil

Method DrawBarScale Class GraphPlus
   LOCAL i
   local nScaleItems := 0
   local nRow := 0, nCol := 0, nData := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nBarWidth := 0
   local nCategoryWidth := 0
   local nScaleItemWidth
   nScaleItems := hmg_len( ::aScale ) 
   if nScaleItems == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nScaleItems - 1 // equal space between categories
   nScaleItemWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCol := ::nLeft
   for i := hmg_len( ::aScale ) to 1 step -1
      BT_DrawText ( ::hDC, ::nBottom, nCol, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
      nCol := nCol + nScaleItemWidth
   next i 
return nil


Method DrawLineGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nFirstCol := 0, nPrevRow := 0, nPrevCol := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   LOCAL i, j, nBarHeight

   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   nFirstCol := ::nLeft + ( nCategoryWidth / 2 )   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for j := 1 to hmg_len( ::aData )
      nCol := nFirstCol
      for i := 1 to hmg_len( ::aData[ 1 ] )
         nData := ::aData[ j, i ]
         nBarHeight := ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange )
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize/2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1 // first point
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow - nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow - nBarHeight
            nPrevCol := nCol
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize/2, nCol - ::nPointSize/2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow + nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow + nBarHeight
            nPrevCol := nCol
         endif
         nCol := nCol + nCategoryWidth
      next i
   next j

return nil

Method DrawPointsGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   LOCAL i, j, nBarHeight

   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange )
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawPieGraph Class GraphPlus
   LOCAL fromrow := 0
   LOCAL fromcol := 0
   LOCAL torow := 0
   LOCAL tocol := 0
   LOCAL topleftrow
   LOCAL topleftcol
   LOCAL toprightrow
   LOCAL toprightcol
   LOCAL bottomrightrow
   LOCAL bottomrightcol
   LOCAL bottomleftrow
   LOCAL bottomleftcol
   LOCAL middletopcol
   LOCAL middleleftrow
   LOCAL middleleftcol
   LOCAL middlebottomcol
   LOCAL middlerightrow
   LOCAL middlerightcol
   LOCAL fromradialrow
   LOCAL fromradialcol
   LOCAL toradialrow
   LOCAL toradialcol
   LOCAL degrees := {}
   LOCAL cumulative := {}
   LOCAL i, sum := 0, lsum := 0
   LOCAL previos_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   local ser_sum := 0, cValue
   
   for i := 1 to hmg_len( ::aData[ 1 ] )
      aadd( aSerieValues, ::aData[ 1, i ] )
   next i   
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 

   fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   
   topleftrow := fromrow
   topleftcol := fromcol
   toprightrow := fromrow
   toprightcol := tocol
   bottomrightrow := torow
   bottomrightcol := tocol
   bottomleftrow := torow
   bottomleftcol := fromcol
   middletopcol := fromcol + int(tocol - fromcol) / 2
   middleleftrow := fromrow + int(torow - fromrow) / 2
   middleleftcol := fromcol
   middlebottomcol := fromcol + int(tocol - fromcol) / 2
   middlerightrow := fromrow + int(torow - fromrow) / 2
   middlerightcol := tocol

   torow := torow + 1
   tocol := tocol + 1
   for i := 1 to HMG_LEN(aSerieValues)
      ser_sum += aSerieValues[i]
   next i
   for i := 1 to HMG_LEN(aSerieValues)
      aadd(degrees,Round(aSerieValues[i]/ser_sum * 360,0))
   next i
   sum := 0
   for i := 1 to HMG_LEN(degrees)
      sum += degrees[i]
   next i
   if sum <> 360
      degrees[HMG_LEN(degrees)] := degrees[HMG_LEN(degrees)] + (360 - sum)
   endif

   sum := 0
   lSum := 0
   for i := 1 to HMG_LEN(degrees)
      sum := sum + degrees[i]
      aadd(cumulative,sum)
      lSum := lSum + ( degrees[i] / 2 )
      aadd( aLabelCumulative, lSum )
      lSum := lSum + ( degrees[i] / 2 )
   next i
   for i := 1 to hmg_len( cumulative )
      if cumulative[ i ] > 90
         cumulative[ i ] := 450 - cumulative[ i ]
      else
         cumulative[ i ] := 90 - cumulative[ i ]
      endif
      if aLabelCumulative[ i ] > 90
         aLabelcumulative[ i ] := 450 - aLabelcumulative[ i ]
      else
         aLabelcumulative[ i ] := 90 - aLabelcumulative[ i ]
      endif
   next i   
      

   previos_cumulative := -1

   fromradialrow := toprightrow
   fromradialcol := middletopcol
   for i := 1 to HMG_LEN(cumulative)

      if cumulative[i] == previos_cumulative
         loop
      endif

      previos_cumulative := cumulative[i]
      do case
      case cumulative[i] <= 45
         toradialcol := middlerightcol
         toradialrow := middlerightrow - Round(cumulative[i] / 45 * (middlerightrow - toprightrow),0)
      case cumulative[i] <= 90 .and. cumulative[i] > 45
         toradialrow := toprightrow
         toradialcol := toprightcol - Round((cumulative[i] - 45) / 45 * (toprightcol - middletopcol),0)
      case cumulative[i] <= 135 .and. cumulative[i] > 90
         toradialrow := topleftrow
         toradialcol := middletopcol - Round((cumulative[i] - 90) / 45 * (middletopcol - topleftcol),0)
      case cumulative[i] <= 180 .and. cumulative[i] > 135
         toradialcol := topleftcol
         toradialrow := topleftrow + Round((cumulative[i] - 135) / 45 * (middleleftrow - topleftrow),0)
      case cumulative[i] <= 225 .and. cumulative[i] > 180
         toradialcol := topleftcol
         toradialrow := middleleftrow + Round((cumulative[i] - 180) / 45 * (bottomleftrow - middleleftrow),0)
      case cumulative[i] <= 270 .and. cumulative[i] > 225
         toradialrow := bottomleftrow
         toradialcol := bottomleftcol + Round((cumulative[i] - 225) / 45 * (middlebottomcol - bottomleftcol),0)
      case cumulative[i] <= 315 .and. cumulative[i] > 270
         toradialrow := bottomleftrow
         toradialcol := middlebottomcol + Round((cumulative[i] - 270) / 45 * (bottomrightcol - middlebottomcol),0)
      case cumulative[i] <= 360 .and. cumulative[i] > 315
         toradialcol := bottomrightcol
         toradialrow := bottomrightrow - Round((cumulative[i] - 315) / 45 * (bottomrightrow - middlerightrow),0)
      endcase
      BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, toradialrow, toradialcol, fromradialrow, fromradialcol, ::aBackColor, ::nPieGap, ::aColors[ i ] )
      fromradialrow := toradialrow
      fromradialcol := toradialcol
      
   next i
   if ::lShowValues
      nChartSize :=  nChartSize / 1.5 

      fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
      fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
      
      topleftrow := fromrow
      topleftcol := fromcol
      toprightrow := fromrow
      toprightcol := tocol
      bottomrightrow := torow
      bottomrightcol := tocol
      bottomleftrow := torow
      bottomleftcol := fromcol
      middletopcol := fromcol + int(tocol - fromcol) / 2
      middleleftrow := fromrow + int(torow - fromrow) / 2
      middleleftcol := fromcol
      middlebottomcol := fromcol + int(tocol - fromcol) / 2
      middlerightrow := fromrow + int(torow - fromrow) / 2
      middlerightcol := tocol

      torow := torow + 1
      tocol := tocol + 1
   
   
      previos_cumulative := -1


      fromradialrow := toprightrow
      fromradialcol := middletopcol
      for i := 1 to HMG_LEN(cumulative)
         if aLabelCumulative[i] == previos_cumulative
            loop
         endif
         cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
         previos_cumulative := aLabelCumulative[i]
         do case
         case aLabelCumulative[i] <= 45
            toradialcol := middlerightcol
            toradialrow := middlerightrow - Round(aLabelCumulative[i] / 45 * (middlerightrow - toprightrow),0)
         case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
            toradialrow := toprightrow
            toradialcol := toprightcol - Round((aLabelCumulative[i] - 45) / 45 * (toprightcol - middletopcol),0)
         case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
            toradialrow := topleftrow
            toradialcol := middletopcol - Round((aLabelCumulative[i] - 90) / 45 * (middletopcol - topleftcol),0)
         case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
            toradialcol := topleftcol
            toradialrow := topleftrow + Round((aLabelCumulative[i] - 135) / 45 * (middleleftrow - topleftrow),0)
         case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
            toradialcol := topleftcol
            toradialrow := middleleftrow + Round((aLabelCumulative[i] - 180) / 45 * (bottomleftrow - middleleftrow),0)
         case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
            toradialrow := bottomleftrow
            toradialcol := bottomleftcol + Round((aLabelCumulative[i] - 225) / 45 * (middlebottomcol - bottomleftcol),0)
         case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
            toradialrow := bottomleftrow
            toradialcol := middlebottomcol + Round((aLabelCumulative[i] - 270) / 45 * (bottomrightcol - middlebottomcol),0)
         case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
            toradialcol := bottomrightcol
            toradialrow := bottomrightrow - Round((aLabelCumulative[i] - 315) / 45 * (bottomrightrow - middlerightrow),0)
         endcase
         BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      next i
   endif
   
return nil

Method DrawBarGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nBarHeight := 0
   local nCategoryHeight := 0
   local nFactor := 0
   local nBarGap := 0
   local i, j
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio 
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
   nCategoryHeight := nBarHeight * ::nGapWidthRatio
   nBarGap      := nBarHeight * ::nBarGapRatio

   nRow := ::nTop + ( nCategoryHeight / 2 )
   nCol := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarWidth := ( ::nRight - ::nLeft ) * ( abs( nData ) / ::nRange )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nCol + nBarWidth - 5, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT )
            endif
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol - nBarWidth, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nCol - nBarWidth + 5, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_LEFT )
            endif
         endif
         nRow := nRow + nBarHeight
         if j < hmg_len( ::aData )
            nRow := nRow + nBarGap
         endif   
      next j
      nRow := nRow + nCategoryHeight
   next i
return nil

Method DrawFunnelGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nBarHeight := 0
   local nCategoryHeight := 0
   local nFactor := 0
   local nBarGap := 0
   local aFunnelData := aclone( ::aData[ 1 ] )
   local nMax := 0, aSize, nWidth, nFromCol, i
   nCategories := hmg_len( aFunnelData )
   // find max value
   nMax := 0
   FOR i := 1 TO HMG_LEN( aFunnelData )
      nMax := Max( aFunnelData[ i ], nMax )
   NEXT i
   if nMax <= 0 
      msgstop( 'Funnel can not be drawn for negative or zero values!' )      
      return nil
   endif
   // title height
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      ::nTop := ::nTop + aSize[ 2 ]
   endif
   if ::lShowCategories 
      ::DrawCategories()
   endif
   if hmg_len( ::cPicture ) == 0
      ::cPicture := replicate( '9', hmg_len( alltrim( str( nMax ) ) ))
   endif   
   
   ::nRight := ::nRight - 10
   ::nLeft := ::nLeft + 5
   ::nTop := ::nTop + 10
   nTotalBars := hmg_len( aFunnelData )
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nTotalBars * 1 // 100%
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
   nCategoryHeight := nBarHeight * ::nGapWidthRatio
   // draw left vertical line
   BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
   nWidth := ( ::nRight - ::nLeft ) - 10
   nRow := ::nTop
   for i := 1 to hmg_len( aFunnelData )
      nData := aFunnelData[ i ]
      nBarWidth := nWidth * ( nData / nMax )
      nFromCol := ::nLeft + 5 + ( ( nWidth - nBarWidth ) / 2 )
      if nData > 0 // ( only for non negative )
         BT_DrawFillRectangle ( ::hDC, nRow, nFromCol, nBarWidth, nBarHeight, ::aColors[ 1 ], ::aColors[ 1 ], 1 )
         BT_DrawText ( ::hDC, ( nRow + nBarHeight / 2 ) , ::nLeft + 5 + ( nWidth / 2 ) , alltrim( transform( aFunnelData[ i ], ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aBackColor, ::aColors[ 1 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_CENTER)
      endif
      nRow := nRow + nBarHeight + nCategoryHeight
   next i
   // title
   if hmg_len( ::cTitle ) > 0
      BT_DrawText ( ::hDC, 5, ::nLeft + 5 + ( nWidth / 2 ), ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
   endif
return nil


Method Save( cFileName ) Class GraphPlus
   if pcount() == 0
      cFileName := GetFile ()
   endif
   cFileName := alltrim( cFileName )
   if hmg_len( cFileName ) > 0   
      if file( cFileName ) 
         if .not. msgyesno( 'File already Exists! Do you want to overwrite?' )
            return nil
         endif
      endif
      BT_BitmapSaveFile( ::hBitMap, cFileName, BT_FILEFORMAT_PNG )
   endif      
return nil


Method DrawLegends Class GraphPlus
   local nMaxWidth := 0, i := 0, j, nMaxHeight := 0
   local nLegends := hmg_len( ::aLegends )
   local aSize := {}
   local nColumns := 0, nRows := 0
   local nRow := 0
   local nCol := 0
   for i := 1 to nLegends
      aSize := BT_DrawTextSize (::hDC, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
   next i   
   nMaxWidth := ( nMaxWidth + ::nLegendBoxSize + 10 ) // per legend
   nMaxHeight := nMaxHeight + 10                      // per legend
   ::nLegendWidth := nMaxWidth
   
   // find columns/rows
   if ::nLegendPos == 1 // bottom
      nColumns := int( ( ::nWidth -  20 ) / nMaxWidth )
      if nColumns == 0 // legend length is not fitting!
         nColumns := 1
      endif
      nRows := int( nLegends / nColumns ) + ( if( mod( nLegends, nColumns ) > 0, 1, 0 ) )
   else // right
      nColumns := 1
      nRows := nLegends
      if nLegends * nMaxHeight > ( ::nBottom - ::nTop )
         // all the legends can not fit!
         nRows := int( ( ::nBottom - ::nTop ) / nMaxHeight )
      endif
   endif   
   // draw legends
   if ::nLegendPos == 1 // bottom
      ::nBottom := ::nBottom - ( nRows * nMaxHeight )
      nRow := ::nBottom + 5
      i := 1
      j := 1
      nCol := ( ::nWidth - nLegends * ::nLegendWidth ) / 2
      do while i <= nLegends
         if j > nColumns
            j := 1
            nRow := nRow + nMaxHeight
            nCol := ( ::nWidth - nLegends * ::nLegendWidth ) / 2
         endif
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nCol := nCol + nMaxWidth
         j++
         i++
      enddo
   else  // right
      ::nRight := ::nRight - nMaxWidth
      nRow := ::nTop + 5
      nCol := ::nRight + 5
      for i := 1 to nRows
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nRow := nRow + nMaxHeight         
      next i   
   endif   
return nil

Method DrawCategories Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0, nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0
   local aSize := {}, nMaxWidth := 0, nMaxHeight, nCategoryHeight, i

   local nBarWidth := 0, nBarHeight := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   nCategories := hmg_len( ::aCategories )
   if nCategories == 0
      return nil
   endif
   if .not. ::lGraphRotated // bar
      nAvailableWidth := ::nRight - ( ::nLeft + ::nMaxScaleWidth + 5 )
      nTotalBarRatio := nCategories // equal space between categories

      nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
      nCategoryWidth := nBarWidth 
      
      nCol := ::nLeft + ::nMaxScaleWidth + 5 + ( nCategoryWidth / 2 )
      nRow := ::nZeroPos
      nMaxHeight := 0
      asize := {}
      for i := 1 to hmg_len( ::aCategories )
         aSize := BT_DrawTextSize (::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
      next i
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawText ( ::hDC, ::nBottom, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
         nCol := nCol + nCategoryWidth
      next i
      ::nBottom := ::nBottom - nMaxHeight
   else
      nAvailableHeight := ::nBottom - ::nTop
      nTotalBarRatio := nCategories // equal space between categories

      nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
      nCategoryHeight := nBarHeight
      
      nRow := ::nTop + ( nCategoryHeight / 2 )
      nMaxWidth := 0
      asize := {}
      for i := 1 to hmg_len( ::aCategories )
         aSize := BT_DrawTextSize (::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      next i
      nCol := ::nLeft + nMaxWidth
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawText ( ::hDC, nRow + 2, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT)
         nRow := nRow + nCategoryHeight
      next i
      ::nLeft := ::nLeft + nMaxWidth
      ::nLeft := ::nLeft + 5
   endif
return nil

Method DrawXYTitles Class GraphPlus
   local aSize := {}
   local cXTitle := ::cXTitle
   local cYTitle := ::cYTitle
   if ::lGraphRotated // bar graph -> Swap x and y
      cYTitle := ::cXTitle
      cXTitle := ::cYTitle
   endif      
   if hmg_len( cXTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, ::nBottom - aSize[ 2 ], (::nRight - ::nLeft) / 2, cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
      ::nBottom := ::nBottom - aSize[ 2 ]
   endif
   if hmg_len( cYTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, (::nBottom - ::nTop) / 2, ::nLeft, cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER, BT_TEXT_VERTICAL_ASCENDANT)
      ::nLeft := ::nLeft + aSize[ 2 ]
   endif
return nil

Method InitColors Class GraphPlus
   local i, j, nLegends, nFactor, nCurFactor := 10
   local aHSL
   local aColorTheme
   nLegends := hmg_len( ::aLegends )
   if hmg_len( ::aColors ) = 0 .or. hmg_len( ::aColors ) < nLegends
      asize( ::aColors, 0 )
      aColorTheme := ::aColorThemes[ ::nColorTheme ]
      if nLegends <= hmg_len( aColorTheme )
         for i := 1 to nLegends
            aadd( ::aColors, aColorTheme[ i ] )
         next i
      else
         for i := 1 to hmg_len( aColorTheme )
            aadd( ::aColors, aColorTheme[ i ] )         
         next i
         i := hmg_len( aColorTheme ) + 1
         j := 1
         nFactor := 10
         do while i <= hmg_len( ::aLegends )
            if j > hmg_len( aColorTheme )
               j := 1
               nCurFactor *= nFactor
            endif   
            aHSL := GP_RGB2HSL( aColorTheme[ j ] )
            aHSL[ 3 ] := aHSL[ 3 ] + iif( ::lLighter, nCurFactor, -nCurFactor )
            aadd( ::aColors, GP_HSL2RGB( aHSL ) )
            j++
            i++
         enddo
      endif
   endif
return nil

Method SwitchRowsToColumns
   local aTemp := {}, aLine := {}
   local cTemp := ''
   local i := 0
   local j := 0
   // transpose data
   for i := 1 to hmg_len( ::aData[ 1 ] )
      asize( aLine, 0 )
      for j := 1 to hmg_len( ::aData )
         aadd( aLine, ::aData[ j, i ] )
      next j
      aadd( aTemp, aclone( aLine ) )
   next i
   asize( ::aData, 0 )   
   ::aData := aclone( aTemp )
   // switch legends and categories
   aTemp := aclone( ::aCategories )
   ::aCategories := aclone( ::aLegends )
   ::aLegends := aclone( aTemp )
   // switch xy titles
   cTemp := ::cXTitle
   ::cXTitle := ::cYTitle
   ::cYTitle := cTemp
return nil   




function findboundary( nNum )
   LOCAL nE, nMax, nMan, nVal, nOffset

   nE:=10
   nVal:=0
   nNum:=Abs(nNum)

   DO WHILE .T.

      nMax := 10**nE

      IF Int(nNum/nMax)>0

         nMan:=(nNum/nMax)-Int(nNum/nMax)
         nOffset:=1
         nOffset:=IF(nMan<=.75,.75,nOffset)
         nOffset:=IF(nMan<=.50,.50,nOffset)
         nOffset:=IF(nMan<=.25,.25,nOffset)
         nOffset:=IF(nMan<=.00,.00,nOffset)
         nVal := (Int(nNum/nMax)+nOffset)*nMax
         EXIT

      ENDIF

      nE--

   ENDDO

RETURN (nVal)
   

function GetNegativeColor( aRGB )
   local aNewRGB := {}
   aNewRGB := { 255 - aRGB[ 1 ], 255 - aRGB[ 2 ], 255 - aRGB[ 3 ] }
return aNewRGB
   
FUNCTION GP_RGB2HSL( aRGB )

   LOCAL nMax, nMin, nDelta
   LOCAL nH, nS, nL, nR, nG, nB

   nR := aRGB[1] / 255
   nG := aRGB[2] / 255
   nB := aRGB[3] / 255
   nMax := Max( nR, Max( nG, nB ) )
   nMin := Min( nR, Min( nG, nB ) )
   nDelta := nMax - nMin

   IF nDelta == 0
      nH := 0
   ELSEIF ( nMax == nR )
      nH := (( nG - nB) / nDelta) % 6
   ELSEIF ( nMax == nG )
      nH := (nB - nR) / nDelta + 2
   ELSE
      nH := (nR - nG) / nDelta + 4
   ENDIF
   nH := Int( nH * 60 )
   IF ( nH < 0 )
      nH += 360
   ENDIF
   nL := ( nMax + nMin ) / 2
   nS := iif( nDelta == 0, 0, nDelta / ( 1 - Abs( 2 * nL - 1 )) )
   nS := nS * 100
   nL := nL * 100
   RETURN { nH, nS, nL }

FUNCTION GP_HSL2RGB( aHSL )

   LOCAL nR := 0, nG := 0, nB := 0, nH, nS, nL, c, x, m
   nH := aHSL[1]
   nS := aHSL[2] / 100
   nL := aHSL[3] / 100
   c := ( 1 - Abs( 2 * nL - 1 ) ) * nS
   x := c * ( 1 - Abs( ( nH / 60 ) % 2 - 1 ) )
   m := nL - c / 2
   IF (0 <= nH .And. nH < 60)
      nR := c
      nG := x
      nB := 0
   ELSEIF (60 <= nH .And. nH < 120)
      nR := x
      nG := c
      nB := 0
   ELSEIF (120 <= nH .And. nH < 180)
      nR := 0
      nG := c
      nB := x
   ELSEIF (180 <= nH .And. nH < 240)
      nR := 0
      nG := x
      nB := c
   ELSEIF (240 <= nH .And. nH < 300)
      nR := x
      nG := 0
      nB := c
   ELSEIF (300 <= nH .And. nH < 360)
      nR := c
      nG := 0
      nB := x
   ENDIF
   nR := Int( ( nR + m ) * 255 )
   nG := Int( ( nG + m ) * 255 )
   nB := Int( ( nB + m ) * 255 )
   RETURN { nR, nG, nB }   
   
function degreetoradian( nDegree )
   return ( nDegree * 0.0174532925 )

function radiantodegree( nRadian )
   return ( nRadian / 0.0174532925 )  

   
   